import unittest
from decimal import Decimal

import mock
from facepy.signed_request import SignedRequestError

from pyramid_facebook.tests import conf
from pyramid_facebook.tests.functional import TestView
from pyramid_facebook.tests.integration import test_payments


class TestRealTimePaymentsUpdate(test_payments.Mixin, unittest.TestCase):

    def setUp(self):
        self.config.add_settings(conf)
        self.config.include('pyramid_facebook')


class TestFacebookPayments(TestView):

    # https://developers.facebook.com/docs/howtos/payments/fulfillment/
    def _sigreq(self):
        return {
            'algorithm': 'HMAC-SHA256',
            'issued_at': 1367617646,
            'payment_id': 370022,
            'quantity': 1,
            'status': 'completed',
        }

    def _payment(self):
        return {
            'id': '370022',
            'user': {
                'id': '123',
                'name': 'Peter',
            },
            'application': {
                'name': 'Yo App',
                'namespace': 'yo_app_tests',
                'id': '555999',
            },
            'actions': [
                {
                    'type': 'charge',
                    'status': 'completed',
                    'currency': 'GBP',
                    'amount': '0.05',
                    'time_created': '2013-05-14T05:35:24+0000',
                    'time_updated': '2013-05-14T05:35:24+0000',
                },
            ],
            'refundable_amount': {
                'currency': 'GBP',
                'amount': '0.05',
            },
            'items': [
                {
                    'type': 'IN_APP_PURCHASE',
                    'product': 'http://example.com/og/pack',
                    'quantity': 1,
                },
            ],
            'country': 'GB',
            'created_time': '2013-05-14T05:35:24+0000',
            'payout_foreign_exchange_rate': Decimal('1.5124725'),
        }

    @mock.patch('requests.get')
    @mock.patch('pyramid_facebook.payments.get_application_access_token')
    @mock.patch('facepy.SignedRequest.parse')
    def test_put_order(self, m_sreq_parse, m_get_app_token, m_http_get):
        m_sreq_parse.return_value = self._sigreq()
        m_http_get.return_value.json.return_value = self._payment()

        self.app.put('/test_canvas/orders/370022',
                     params={'signed_request': 'fake'}, status=200)

    @mock.patch('facepy.SignedRequest.parse')
    def test_put_order_invalid_signed_request(self, m_sreq_parse):
        m_sreq_parse.side_effect = SignedRequestError('Nope')

        self.app.put('/test_canvas/orders/380088',
                     params={'signed_request': 'fake'}, status=400)

    @mock.patch('facepy.SignedRequest.parse')
    def test_put_order_invalid_payment_id(self, m_sreq_parse):
        m_sreq_parse.return_value = {
            'algorithm': 'HMAC-SHA256',
            'issued_at': 1367617646,
            'payment_id': 360066,
            'quantity': 1,
            'status': 'completed',
        }

        self.app.put('/test_canvas/orders/30055',
                     params={'signed_request': 'fake'}, status=400)

    @mock.patch('requests.get')
    @mock.patch('pyramid_facebook.payments.get_application_access_token')
    @mock.patch('facepy.SignedRequest.parse')
    def test_put_order_notify_error(self, m_sreq_parse, m_get_app_token,
                                    m_http_get):
        from pyramid_facebook.events import OrderReceived, OrderProcessingError
        m_sreq_parse.return_value = self._sigreq()
        m_http_get.return_value.json.return_value = self._payment()

        def subscriber_fail(event):
            raise Exception('Shit hits the fan.')

        logs = []

        def subscriber_report(event):
            logs.append(event)

        self.config.add_subscriber(subscriber_fail, OrderReceived)
        self.config.add_subscriber(subscriber_report, OrderProcessingError)

        self.app.put('/test_canvas/orders/370022',
                     params={'signed_request': 'fake'}, status=500)

        self.assertEqual(len(logs), 1, logs)
        self.assertIsInstance(logs[0], OrderProcessingError)

    @mock.patch('requests.get')
    @mock.patch('pyramid_facebook.payments.get_application_access_token')
    @mock.patch('facepy.SignedRequest.parse')
    def test_put_order_double_notify_error(self, m_sreq_parse, m_get_app_token,
                                           m_http_get):
        from pyramid_facebook.events import OrderReceived, OrderProcessingError
        m_sreq_parse.return_value = self._sigreq()
        m_http_get.return_value.json.return_value = self._payment()

        def subscriber_fail(event):
            raise Exception('Shit hits the fan.')

        self.config.add_subscriber(subscriber_fail, OrderReceived)
        self.config.add_subscriber(subscriber_fail, OrderProcessingError)

        self.app.put('/test_canvas/orders/370022',
                     params={'signed_request': 'fake'}, status=500)



import mock

from . import TestView


def dummy_view(request):
    graph_api = request.graph_api
    return {'app_token': graph_api.oauth_token}


class Test(TestView):

    @mock.patch('pyramid_facebook.utility.facepy.get_application_access_token')
    def test_get_application_graph_api(self, m_get_app_token):
        from pyramid_facebook.utility import get_application_graph_api
        m_get_app_token.return_value = 'app_token'

        graph_api1 = get_application_graph_api(self.config.registry)
        graph_api2 = get_application_graph_api(self.config.registry)

        self.assertIs(graph_api1, graph_api2)

    @mock.patch('pyramid_facebook.utility.facepy.get_application_access_token')
    def test_includeme(self, m_get_app_token):
        m_get_app_token.return_value = 'app_token'

        self.config.include('pyramid_facebook.utility')
        self.config.add_route('test', '/test')
        self.config.add_view(dummy_view, route_name='test', renderer='json')

        self.assertEqual(
            {u'app_token': 'app_token'},
            self.app.get('/test').json
            )



from mock import patch

from pyramid_facebook.lib import encrypt_signed_request

__all__ = ['conf', 'get_signed_request']

conf = {
    'facebook.app_id': '1234567890',
    'facebook.secret_key': 'abcdef1234567890',
    'facebook.namespace': 'test_canvas',
    'facebook.scope': 'publish_actions,email',
    'facebook.rt-subscriptions': ('\nuser = friends, name\n'
                                  'payments = actions, disputes'),
    'pyramid.debug_routematch': True,
    }


def get_signed_request(**params):
    return encrypt_signed_request(conf['facebook.secret_key'], params)


def setup_package():
    patch('pyramid_facebook.security.get_application_access_token').start()
    p = patch('pyramid_facebook.utility.facepy.get_application_access_token')
    p.start()



import logging

import facepy

from zope.interface import Interface
from zope.interface.declarations import classImplements

log = logging.getLogger(__name__)


def includeme(config):
    config.add_request_method(request_get_application_graph_api, 'graph_api',
                              reify=True)


def request_get_application_graph_api(request):
    return get_application_graph_api(request.registry)


def get_application_graph_api(registry):
    graph_api = registry.queryUtility(IGraphAPI, name='application_graph_api')

    if graph_api is None:
        log.debug('Create GraphAPI with facebook application token')

        app_id = registry.settings['facebook.app_id']
        secret_key = registry.settings['facebook.secret_key']
        app_token = facepy.get_application_access_token(app_id, secret_key)
        graph_api = facepy.GraphAPI(app_token)
        registry.registerUtility(graph_api, name='application_graph_api')

    return graph_api


class IGraphAPI(Interface):
    pass


classImplements(facepy.GraphAPI, IGraphAPI)



import unittest
import json

import mock


from pyramid import testing
from pyramid.security import Allow

from pyramid_facebook.security import SignedRequestContext
from pyramid_facebook.tests import conf

TEST_USER = 123


def _get_signed_request(user_id=None):
    from pyramid_facebook.lib import encrypt_signed_request
    return encrypt_signed_request(
        conf['facebook.secret_key'],
        {
            'user_id': user_id if user_id else TEST_USER,
            'user': {
                'country': 'ca'
                }
            }
        )


def _get_request(signed_request=None):
    request = mock.MagicMock()
    signed_request = (signed_request
                      if signed_request else _get_signed_request())
    request.params = {'signed_request': signed_request}
    request.registry.settings = conf.copy()
    return request


class TestIncludeme(unittest.TestCase):

    def test_includeme(self):
        from pyramid_facebook.security import includeme

        config = mock.MagicMock()

        includeme(config)


class TestFacebookCreditsContext(unittest.TestCase):

    @mock.patch('pyramid_facebook.security.json')
    def test_order_details(self, m_json):
        from pyramid_facebook.security import FacebookCreditsContext
        ctx = FacebookCreditsContext(_get_request())
        ctx.facebook_data = data = mock.MagicMock()

        self.assertEqual(m_json.loads.return_value, ctx.order_details)
        m_json.loads.assert_called_once_with(data['credits']['order_details'])

    def test_earned_currency_data(self):
        from pyramid_facebook.security import FacebookCreditsContext
        ctx = FacebookCreditsContext(_get_request())
        ctx.facebook_data = {
            'credits': {
                'order_details': json.dumps({
                    'items': [
                        {'data': '{"modified": "earned_currency_data"}'}
                        ]
                    })
                }
            }
        self.assertEqual(
            'earned_currency_data',
            ctx.earned_currency_data
            )

    def test_earned_currency_data_invalid_json(self):
        from pyramid_facebook.security import FacebookCreditsContext
        ctx = FacebookCreditsContext(_get_request())
        # with Invalid json data:
        ctx.facebook_data = {
            'credits': {
                'order_details': json.dumps({
                    'items': [
                        {'data': '{INVALID JSON "}'}
                        ]
                    })
                }
            }
        self.assertEqual(
            None,
            ctx.earned_currency_data
            )

    def test_order_info(self):
        from pyramid_facebook.security import FacebookCreditsContext
        ctx = FacebookCreditsContext(_get_request())
        ctx.facebook_data = mock.MagicMock()

        self.assertEqual(ctx.facebook_data["credits"]["order_info"],
                         ctx.order_info)

    def test_item(self):
        from pyramid_facebook.security import FacebookCreditsContext
        ctx = FacebookCreditsContext(_get_request())
        ctx.facebook_data = mock.MagicMock()
        t = json.dumps({'items': [{'title': 'an item'}]})
        ctx.facebook_data.__getitem__.return_value.__getitem__.return_value = t

        self.assertEqual(
            {'title': 'an item'},
            ctx.item,
            )


class TestSignedRequestContext(unittest.TestCase):

    def test_init(self):
        from pyramid_facebook.security import (
            SignedRequestContext,
            ViewCanvas,
            Authenticate,
            FacebookUser,
            RegisteredUser,
            )

        self.assertEqual(
            [
                (Allow, FacebookUser, Authenticate),
                (Allow, RegisteredUser, ViewCanvas),
                ],
            SignedRequestContext.__acl__
            )

        SignedRequestContext(None)

    def test_facebook_data_not_set(self):
        from pyramid_facebook.security import SignedRequestContext
        sr = SignedRequestContext(None)
        self.assertEqual(None, sr.facebook_data)

    def test_facebook_data_property(self):
        from pyramid_facebook.security import SignedRequestContext
        sr = SignedRequestContext(None)
        sr.facebook_data = 'something'
        self.assertEqual('something', sr.facebook_data)

    def test_repr(self):
        from pyramid_facebook.security import SignedRequestContext

        ctx = SignedRequestContext(_get_request())
        # just make sure that repr works, the exact format is not guaranteed
        repr(ctx)

    def test_user(self):
        from pyramid_facebook.security import SignedRequestContext
        sr = SignedRequestContext(None)
        sr.facebook_data = {'user': 12345}
        self.assertEqual(12345, sr.user)

    def test_country(self):
        from pyramid_facebook.security import SignedRequestContext
        sr = SignedRequestContext(None)
        sr.facebook_data = {'user': {'country': 'FR'}}
        self.assertEqual('FR', sr.user_country)


class TestAccessTokenContext(unittest.TestCase):

    def test_access_token(self):
        from pyramid_facebook.security import AccessTokenContext
        request = mock.Mock()
        atc = AccessTokenContext(request)
        atc.user_dict = {u'k': 123}
        atc.user_id = 1000

        self.assertEqual(atc.request, request)
        self.assertEqual({u'k': 123}, atc.user_dict)
        self.assertEqual(1000, atc.user_id)


class TestRealTimeNotificationContext(unittest.TestCase):

    def test_init(self):
        from pyramid_facebook.security import (
            RealTimeNotificationContext,
            Allow,
            XHubSigned,
            NotifyRealTimeChanges
            )

        request = mock.Mock()

        rtnc = RealTimeNotificationContext(request)

        self.assertEqual(request, rtnc.request)
        self.assertEqual(
            ((Allow, XHubSigned, NotifyRealTimeChanges), ),
            rtnc.__acl__
            )


class TestAuthPolicy(unittest.TestCase):

    def setUp(self):
        self.config = testing.setUp()

    def tearDown(self):
        testing.tearDown()

    def _get_request(self, context_cls=SignedRequestContext):
        request = testing.DummyRequest()
        request.params = {'signed_request': mock.MagicMock(spec=str)}
        request.registry.settings = {
            'facebook.secret_key': 'a secret',
            }
        request.context = mock.MagicMock(spec=context_cls)
        return request


class TestSignedRequestAuthenticationPolicy(TestAuthPolicy):

    @mock.patch('pyramid_facebook.security.SignedRequest')
    def test_unauthenticated_userid(self, SignedRequest):
        from pyramid_facebook.security import SignedRequestAuthenticationPolicy
        request = self._get_request()
        SignedRequest.parse.return_value = {u'user_id': 123}

        result = (SignedRequestAuthenticationPolicy()
                  .unauthenticated_userid(request))

        SignedRequest.parse.assert_called_once_with(
            request.params['signed_request'],
            request.registry.settings['facebook.secret_key'],
            )
        self.assertEqual(123, result)

    def test_unauthenticated_userid_with_no_signed_request(self):
        from pyramid_facebook.security import SignedRequestAuthenticationPolicy
        request = self._get_request()
        del request.params['signed_request']

        self.assertIsNone(
            SignedRequestAuthenticationPolicy().unauthenticated_userid(request)
            )

    @mock.patch('pyramid_facebook.security.SignedRequest')
    def test_unauthenticated_userid_with_no_user_id(self, SignedRequest):
        from pyramid_facebook.security import SignedRequestAuthenticationPolicy
        request = self._get_request()
        SignedRequest.parse.return_value = {'no_user_id': ':-o'}

        self.assertIsNone(
            SignedRequestAuthenticationPolicy().unauthenticated_userid(request)
            )

    @mock.patch('pyramid_facebook.security.SignedRequest')
    def test_unauthenticated_userid_with_invalid_user_id(self, SignedRequest):
        from pyramid_facebook.security import SignedRequestAuthenticationPolicy
        request = self._get_request()
        SignedRequest.parse.return_value = {u'user_id': u'not an int'}

        self.assertIsNone(
            SignedRequestAuthenticationPolicy().unauthenticated_userid(request)
            )

    def test_unauthenticated_userid_with_bad_signed_request(self):
        from pyramid_facebook.security import SignedRequestAuthenticationPolicy
        request = self._get_request()
        request.params = {u'signed_request': u'wrong sigmed request'}
        self.assertIsNone(
            SignedRequestAuthenticationPolicy().unauthenticated_userid(request)
            )

    def test_effective_principals_no_facebook_data(self):
        from pyramid_facebook.security import SignedRequestAuthenticationPolicy
        request = self._get_request()
        request.context.facebook_data = {}
        self.assertEqual(
            [],
            SignedRequestAuthenticationPolicy().effective_principals(request)
            )


class TestAccessTokenAuthenticationPolicy(TestAuthPolicy):

    @mock.patch('pyramid_facebook.security.GraphAPI')
    def test_validate_access_token_with_no_access_token(self, GraphAPI):
        from pyramid_facebook.security import AccessTokenAuthenticationPolicy
        request = self._get_request()

        self.assertIsNone(
            AccessTokenAuthenticationPolicy().unauthenticated_userid(request)
            )

    @mock.patch('pyramid_facebook.security.GraphAPI')
    def test_validate_access_token_with_facepy_error(self, GraphAPI):
        from pyramid_facebook.security import (
            AccessTokenAuthenticationPolicy,
            FacepyError,
            AccessTokenContext,
            )
        request = self._get_request(context_cls=AccessTokenContext)
        request.context.user_dict = {}
        request.params = {u'access_token': u'123qwerty'}

        api = GraphAPI.return_value = mock.Mock()
        api.get.side_effect = FacepyError('oooops')

        self.assertIsNone(
            AccessTokenAuthenticationPolicy().unauthenticated_userid(request)
            )

    @mock.patch('pyramid_facebook.security.GraphAPI')
    def test_validate_access_token_malformed_fb_response(self, GraphAPI):
        from pyramid_facebook.security import (
            AccessTokenAuthenticationPolicy,
            AccessTokenContext,
            )
        request = self._get_request(context_cls=AccessTokenContext)
        request.params = {u'access_token': u'123qwerty'}

        api = GraphAPI.return_value = mock.Mock()
        api.get.return_value = {u'id': u'not an int'}

        self.assertIsNone(
            AccessTokenAuthenticationPolicy().unauthenticated_userid(request)
            )

    @mock.patch('pyramid_facebook.security.GraphAPI')
    def test_validate_access_token(self, GraphAPI):
        from pyramid_facebook.security import (
            AccessTokenAuthenticationPolicy,
            AccessTokenContext,
            )
        request = self._get_request(context_cls=AccessTokenContext)
        request.params = {u'access_token': u'123qwerty'}

        api = GraphAPI.return_value = mock.Mock()
        api.get.return_value = {u'id': u'12345'}

        self.assertEqual(
            12345,
            AccessTokenAuthenticationPolicy().unauthenticated_userid(request)
            )

    def test_get_principals_from_access_token(self):
        from pyramid_facebook.security import (
            AccessTokenAuthenticationPolicy,
            AccessTokenContext,
            FacebookUser
            )
        request = self._get_request(context_cls=AccessTokenContext)
        request.context.user_id = 1234

        self.assertEqual(
            [FacebookUser],
            AccessTokenAuthenticationPolicy().effective_principals(request)
            )


@mock.patch('pyramid_facebook.security.get_application_access_token')
@mock.patch('pyramid_facebook.security.GraphAPI')
class TestAdminAuthenticationPolicy(TestAuthPolicy):

    def _get_inst(self):
        from pyramid_facebook.security import AdminAuthenticationPolicy
        return AdminAuthenticationPolicy('app_id', 'secret')

    def test_get_principals_from_admin_context_with_facepy_error(
        self,
        GraphAPI,
        get_app_token):
        from pyramid_facebook.security import AdminContext, FacepyError
        request = self._get_request(context_cls=AdminContext)
        request.params = {u'access_token': u'123qwerty'}

        api = GraphAPI.return_value = mock.Mock()
        api.get.side_effect = FacepyError('opps')

        self.assertEqual(
            ['facebook-user'],
            self._get_inst().effective_principals(request),
            )

    def test_get_principals_from_admin_context(self, GraphAPI, get_app_token):
        from pyramid_facebook.security import AdminContext
        request = self._get_request(context_cls=AdminContext)
        request.context.user_id = 1234
        request.params = {u'access_token': u'123qwerty'}

        api = GraphAPI.return_value = mock.Mock()
        api.get.return_value = {
            u'data': [{
                u'user': u'1234',
                u'role':u'administrator',
                u'app_id': u'3445'
                }]
            }

        self.assertEqual(
            [u'facebook-user', u'group:administrator'],
            self._get_inst().effective_principals(request)
            )


class TestRealTimeNotifAuthenticationPolicy(TestAuthPolicy):

    @mock.patch('pyramid_facebook.security.hmac')
    def test_get_principals_from_signature_wrong_signature(self, hmac):
        from pyramid_facebook.security import (
            RealTimeNotifAuthenticationPolicy,
            RealTimeNotificationContext,
            )
        request = self._get_request(context_cls=RealTimeNotificationContext)
        request.headers[u'X-Hub-Signature'] = u'sha1=signature'

        hmac.new.return_value.hexdigest.return_value = u'WRONG'

        self.assertEqual(
            [],
            RealTimeNotifAuthenticationPolicy().effective_principals(request)
            )

    @mock.patch('pyramid_facebook.security.hmac')
    def test_get_principals_from_signature(self, hmac):
        from pyramid_facebook.security import (
            RealTimeNotifAuthenticationPolicy,
            RealTimeNotificationContext,
            XHubSigned,
            )
        request = self._get_request(context_cls=RealTimeNotificationContext)
        request.headers[u'X-Hub-Signature'] = u'sha1=signature'

        hmac.new.return_value.hexdigest.return_value = u'signature'

        self.assertEqual(
            [XHubSigned],
            RealTimeNotifAuthenticationPolicy().effective_principals(request)
            )



from pyramid.security import has_permission


def includeme(config):
    config.add_subscriber_predicate('matched_route',
                                    MatchedRouteEventPredicate)
    config.add_subscriber_predicate('expected_context', ContextEventPredicate)
    config.add_subscriber_predicate('permission', PermissionEventPredicate)
    config.add_subscriber_predicate('update_object', ChangeNotificationObject)


def request_params_predicate(*required_param_keys, **required_params):
    """Custom predicates to check if required parameter are in request
    parameters. Read :ref:`custom route predicates <pyramid:
    custom_route_predicates>`
    for more info::

        # /example?param1&param2=321
        config.add_route(
            'example',
            '/example',
            custom_predicates=[request_params_predicate('param1', param2=321)]
            )
    """
    required = set(required_param_keys)

    def predicate(info, request):
        if not required.issubset(set(request.params)):
            return False

        if any((k, v) for k, v in required_params.iteritems()
               if v != request.params[k]):
                return False
        return True
    return predicate


def nor_predicate(**params):
    """Custom predicate which checks if a parameter is present with possible
    values being one in list values.

    :param **params: parameter names and values list::

        nor_predicate(param_name=(value1, value2))

    """
    names = set(params.keys())

    def predicate(info, request):
        if not names.issubset(request.params):
            return False
        if any(request.params[n]
               for n in names if request.params[n] not in params[n]):
            return False
        return True
    return predicate


def headers_predicate(*header_names, **headers):
    """Custom predicate which check that `header_names` and  `headers`
    name/value pairs are in `request.headers`.
    """
    def predicate(info, request):
        if any(header for header in header_names
               if header not in request.headers):
            return False
        if any((k, v) for k, v in headers.iteritems()
               if k not in request.headers or request.headers[k] != v):
            return False
        return True
    return predicate


class MatchedRouteEventPredicate(object):

    def __init__(self, route_to_match, config):
        self.route_to_match = route_to_match

    def text(self):
        return u'route_to_match = %s' % self.route_to_match

    # predicate hash
    phash = text

    def __call__(self, event):
        try:
            return event.request.matched_route.name == self.route_to_match
        except AttributeError:  # pragma: no cover
            return False


class ContextEventPredicate(object):
    """Filter a request based event subscriber on context class.
    Only works with request based event which have request as an attribute.
    """

    def __init__(self, expected_context_cls, config):
        self.expected_context_cls = expected_context_cls

    def text(self):
        return u'expected_context_cls = %r' % self.expected_context_cls

    # predicate hash
    phash = text

    def __call__(self, event):
        try:
            return isinstance(event.request.context, self.expected_context_cls)
        except AttributeError:  # pragma: no cover
            return False


class PermissionEventPredicate(object):
    """Filter a request based event subscriber with a permission check.
    Only works with request based event which have request as an attribute.

    :param permission: Permission to check.
    """

    def __init__(self, permission, config):
        self.permission = permission

    def text(self):
        return u'permission = %r' % self.permission

    # predicate hash
    phash = text

    def __call__(self, event):
        try:
            access = has_permission(self.permission, event.request.context,
                                    event.request)
            return access.boolval
        except AttributeError:  # pragma: no cover
            return False


class ChangeNotificationObject(object):
    """Filter a :class:`~pyramid_facebook.events.ChangeNotification` event on
    its notification object (payments, user...).
    """
    def __init__(self, obj, config):
        self.object = obj

    def text(self):
        return 'event.object = %s' % self.object

    phash = text

    def __call__(self, event):
        return event.object == self.object



# -*- coding: utf-8 -*-
import sys
import argparse
import logging
import urlparse
import json

from facepy import FacepyError, get_application_access_token, GraphAPI

from pyramid.httpexceptions import (
    HTTPForbidden,
    HTTPOk,
    HTTPInternalServerError,
    )
from pyramid.paster import bootstrap
from pyramid.request import Request
from pyramid.response import Response
from pyramid.view import view_config

from pyramid_facebook.events import ChangeNotification
from pyramid_facebook.predicates import (
    headers_predicate,
    request_params_predicate,
    nor_predicate,
    )
from pyramid_facebook.security import UpdateSubscription, NotifyRealTimeChanges

log = logging.getLogger(__name__)


def includeme(config):
    """Adds routes related to real time updates.

    Routes Added:

    * ``facebook_real_time_subscription_verification`` associated to url
        /{namespace}/real-time

    * ``facebook_real_time_notification`` associated to url
        /{namespace}/real-time

    * ``facebook_real_time_subscriptions`` associated to url
        /{namespace}/real-time/subscriptions

    * ``facebook_real_time_subscriptions_update`` associated to url
        /{namespace}/real-time/subscriptions

    * ``facebook_real_time_subscriptions_delete`` associated to url
        /{namespace}/real-time/subscriptions
    """
    log.debug('Adding route "facebook_real_time_subscription_verification".')
    config.include('pyramid_facebook.security')
    config.add_route(
        'facebook_real_time_subscription_verification',
        '/real-time',
        request_method='GET',
        custom_predicates=[
            request_params_predicate(
                'hub.challenge',
                'hub.verify_token',
                **{'hub.mode': 'subscribe'}
                ),
            ]
        )

    log.debug('Adding route "facebook_real_time_notification".')
    config.add_route(
        'facebook_real_time_notification',
        '/real-time',
        request_method='POST',
        factory='pyramid_facebook.security.RealTimeNotificationContext',
        custom_predicates=[
            headers_predicate(
                'X-Hub-Signature',
                **{'Content-Type': 'application/json'}
                )
            ]
        )

    log.debug('Adding route "facebook_real_time_subscriptions".')
    config.add_route(
        'facebook_real_time_subscriptions',
        '/real-time/subscriptions',
        request_method='GET',
        request_param='access_token',
        factory='pyramid_facebook.security.AdminContext',
        )

    log.debug('Adding route "facebook_real_time_subscriptions_update".')
    config.add_route(
        'facebook_real_time_subscriptions_update',
        '/real-time/subscriptions',
        request_method='POST',
        custom_predicates=[
            nor_predicate(object=('user', 'permissions', 'page', 'errors',
                                  'payment_subscriptions', 'payments',)),
            request_params_predicate('fields', 'access_token'),
            ],
        factory='pyramid_facebook.security.AdminContext',
        )

    log.debug('Adding route "facebook_real_time_subscriptions_delete".')
    config.add_route(
        'facebook_real_time_subscriptions_delete',
        '/real-time/subscriptions',
        request_method='DELETE',
        request_param='access_token',
        factory='pyramid_facebook.security.AdminContext',
        )

    config.scan(package='pyramid_facebook.real_time')


@view_config(route_name='facebook_real_time_subscription_verification')
def verify_real_time_subscription(request):
    verify_token = request.params['hub.verify_token']
    settings = request.registry.settings
    try:
        access_token = get_application_access_token(
            settings['facebook.app_id'],
            settings['facebook.secret_key'],
            )
        if access_token == verify_token:
            return Response(request.params['hub.challenge'])
    except FacepyError:
        log.exception('get_application_access_token failed')
    raise HTTPForbidden()


@view_config(context=FacepyError, renderer='json')
def render_facepy_error(exc, request):
    return dict(
        error=dict(
            type=exc.__class__.__name__,
            code=getattr(exc, 'code'),
            message=exc.message,
            )
        )


@view_config(
    route_name='facebook_real_time_subscriptions',
    renderer='json',
    permission=UpdateSubscription,
    )
def list_real_time_subscriptions(request):
    settings = request.registry.settings
    app_id = settings['facebook.app_id']
    secret_key = settings['facebook.secret_key']
    access_token = get_application_access_token(app_id, secret_key)
    return GraphAPI(access_token).get('%s/subscriptions' % app_id)


@view_config(
    route_name='facebook_real_time_subscriptions_update',
    renderer='json',
    permission=UpdateSubscription,
    )
def update_real_time_subscription(request):
    settings = request.registry.settings
    fb_object = request.params['object']
    fields = request.params['fields']
    access_token = get_application_access_token(
        settings['facebook.app_id'],
        settings['facebook.secret_key'],
        )
    # url of route named `facebook_real_time_subscription_verification`
    # is the same as route named `facebook_real_time_notification` but with
    # different predicates
    url = request.route_url('facebook_real_time_subscription_verification')
    return (GraphAPI(access_token).post(
        '%s/subscriptions' % settings['facebook.app_id'],
        object=fb_object,
        fields=fields,
        callback_url=url,
        verify_token=access_token
        ))


@view_config(
    route_name='facebook_real_time_subscriptions_delete',
    renderer='json',
    permission=UpdateSubscription,
    )
def delete_real_time_subscription(request):
    settings = request.registry.settings
    access_token = get_application_access_token(
        settings['facebook.app_id'],
        settings['facebook.secret_key'],
        )
    return GraphAPI(access_token).delete(
        '%s/subscriptions' % settings['facebook.app_id'],
        )


@view_config(
    route_name='facebook_real_time_notification',
    permission=NotifyRealTimeChanges,
    )
def process_real_time_notification(context, request):
    log.debug('process_real_time_notification json_body=%s', request.json_body)
    try:
        request.registry.notify(ChangeNotification(context, request))
    except Exception:
        log.exception(
            'process_real_time_notification - json_body: %s',
            request.json_body,
            )
        # raise an HTTP error so facebook will send back same update later
        raise HTTPInternalServerError()
    return HTTPOk()


@view_config(route_name='facebook_real_time_notification',
             context=HTTPForbidden)
def forbid_real_time_update(context, request):
    raise HTTPForbidden()


class ManageSubscriptions(object):
    """Manage real-time subscriptions.

    The easiest way to setup real time subscription is to define the
    `facebook.rt-subscriptions` setting in the ini file::

        facebook.rt-subscriptions =
            user = name, friends
            payments = actions, disputes


    Then run the setup command::

        $ pfacebook-real-time env.ini setup


    To list subscriptions::

        $ pfacebook-real-time env.ini list


    To update or create subscriptions without changing the config file::

        $ pfacebook-real-time env.ini update user=friends,name payments=actions


    To delete all subscriptions::

        $ pfacebook-real-time env.ini delete
    """
    def __init__(self, argv):
        self.setup_parsers()
        self.args = self.parser.parse_args(argv)

        self.env = bootstrap(self.args.config_uri)
        self.settings = self.env['registry'].settings
        self.request = self.env['request']

        self.app_id = self.settings['facebook.app_id']
        secret_key = self.settings['facebook.secret_key']

        self.graph_api = GraphAPI(get_application_access_token(self.app_id,
                                                               secret_key))
        self.base_host = self.get_application_base_host(self.graph_api,
                                                        self.app_id)

    def setup_parsers(self):
        self.parser = argparse.ArgumentParser(
            prog='pfacebook-real-time',
            description=self.__doc__,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        self.parser.add_argument(
            'config_uri',
            help='Paste config_uri string expecting the format inifile#name',
            )

        subparsers = self.parser.add_subparsers()

        setup_parser = subparsers.add_parser('setup')
        setup_parser.set_defaults(func=self.setup)

        list_parser = subparsers.add_parser('list')
        list_parser.set_defaults(func=self.list)

        update_parser = subparsers.add_parser('update')
        update_parser.set_defaults(func=self.update)
        update_parser.add_argument('rt_settings', nargs='*',
                                   metavar='object=field1,field2',
                                   help='Objects and fields to update '
                                        'subscription on.')

        delete_parser = subparsers.add_parser('delete')
        delete_parser.set_defaults(func=self.delete)

    def run(self):
        self.args.func()

    def setup(self):
        rt_settings = self.settings['facebook.rt-subscriptions'].split('\n')
        self.do_update_requests(rt_settings)
        self.list()

    def list(self):
        request = self.new_request()
        result = list_real_time_subscriptions(request)
        print json.dumps(result, indent=4)

    def update(self):
        self.do_update_requests(self.args.rt_settings)
        self.list()

    def delete(self):
        request = self.new_request('facebook_real_time_subscriptions_delete')
        delete_real_time_subscription(request)
        self.list()

    def get_application_base_host(self, graph_api, app_id):
        canvas_url = graph_api.get(app_id, fields='canvas_url')['canvas_url']
        return urlparse.urlparse(canvas_url).netloc

    def new_request(self, route_name=None, **kw):
        path = self.request.route_path(route_name) if route_name else ''
        request = Request.blank(path, POST=kw, environ=self.request.environ,
                                host=self.base_host)
        request.registry = self.request.registry
        return request

    def do_update_requests(self, rt_settings):
        for setting in rt_settings:
            setting = setting.replace(' ', '')
            if not setting:
                continue
            obj, fields = setting.split('=')
            req = self.new_request('facebook_real_time_subscriptions_update',
                                   object=obj, fields=fields)
            update_real_time_subscription(req)


def command_line_script():
    ManageSubscriptions(sys.argv[1:]).run()



from zope.interface import implementer

from pyramid_facebook.tests.functional import TestView


class TestOpenGraph(TestView):

    def setUp(self):
        self.define_virtual_goods()

    def define_virtual_goods(self):
        from pyramid_facebook import opengraph

        class PackCollection(object):

            __name__ = None
            __parent__ = None

            def __init__(self, request):
                self.children = {
                    'rainbows_pack': RainbowsCurrencyPack(self),
                }

            def __getitem__(self, key):
                return self.children[key]

        @implementer(opengraph.IResource)
        class RainbowsCurrencyPack(object):

            def __init__(self, parent):
                self.__name__ = 'rainbows_pack'
                self.__parent__ = parent

                self.prefixes = {
                    'product': 'http://ogp.me/ns/product#',
                }
                self.properties = {
                    'og:type': 'og:product',
                    'og:title': 'Rainbows Pack',
                    'og:description': 'Use Rainbows to play more',
                    'product:price': [
                        {'amount': '0.10', 'currency': 'CAD'},
                        {'amount': '0.15', 'currency': 'USD'},
                    ],
                }

        self.config.add_opengraph_collection(PackCollection, 'store/packs')

    def test_basic(self):
        resp = self.app.get('/test_canvas/store/packs/rainbows_pack',
                            status=200)

        lines = resp.body.splitlines()
        self.assertEqual(
            ['<head prefix="og: http://ogp.me/ns#',
             'fb: http://ogp.me/ns/fb#',
             'product: http://ogp.me/ns/product#',
             '">'],
            [line.strip() for line in lines[:4]])
        self.assertEqual('</head>', lines[-1])

        og_url = 'http://localhost/test_canvas/store/packs/rainbows_pack'

        for prop, expected in [
            ('fb:app_id', '1234567890'),
            ('og:type', 'og:product'),
            ('og:url', og_url),
            ('og:title', 'Rainbows Pack'),
            ('og:description', 'Use Rainbows to play more'),
          ]:
            value = resp.html.head.find('meta', property=prop)['content']
            self.assertEqual(expected, value)

    def test_url_with_query_string(self):
        resp = self.app.get(
            '/test_canvas/store/packs/rainbows_pack?promo=labor_day',
            status=200)

        self.assertEqual(
            resp.html.head.find('meta', property='og:url')['content'],
            'http://localhost/test_canvas/store/packs/rainbows_pack'
            '?promo=labor_day')

    def test_unknown_pack(self):
        self.app.get('/test_canvas/store/packs/unicorns_pack', status=404)

    def test_default_view(self):
        self.app.put('/test_canvas/store/packs/rainbows_pack', status=405)



import json
import logging
from decimal import Decimal

import requests
from facepy import SignedRequest, FacepyError
from facepy.utils import get_application_access_token
from facepy.signed_request import SignedRequestError
from pyramid.events import subscriber
from pyramid.httpexceptions import (
    HTTPBadRequest,
    HTTPFailedDependency,
    HTTPServerError,
)
from pyramid.view import view_config

from pyramid_facebook.events import (
    OrderReceived,
    OrderProcessingError,
    ChangeNotification,
)

log = logging.getLogger(__name__)


def includeme(config):
    """Add views related to local currency payments."""
    config.include('pyramid_facebook.security')
    config.include('pyramid_facebook.predicates')
    config.include('pyramid_facebook.utility')
    config.add_route('facebook_payments_order',
                     'orders/{order_id}')
    config.scan('.payments')


@view_config(route_name='facebook_payments_order',
             request_method='PUT',
             request_param='signed_request',
             renderer='string')
def put_order(context, request):
    """Use this view as server-side verification method.

    If you want to verify payments and maybe update user
    info on the backend side, write a client-side callback
    that sends the payment's signed request data (effectively
    keeping the old model of a blocking credits callback
    instead of using a real-time update handler).

    This view checks the payment using the Graph API and
    sends an OrderReceived event.  If a subscriber raises an
    exception, it is logged, OrderProcessingError is sent (an
    exception in one of its subscribers will also be logged),
    and HTTPServerError is raised.  An event subscriber doing
    something non-critical like tracking should catch all its
    exceptions, otherwise the client code will think the order
    was not processed.

    If the signed request is invalid, or if the payment ID in
    the signed request does not match the current URI,
    HTTPBadRequest will be raised and no event will be sent.
    This indicates a configuration or application bug; the
    error will be logged to aid debugging and refunding.

    See https://developers.facebook.com/docs/howtos/payments/fulfillment/
    """
    signed_request = request.POST['signed_request']
    app_id = request.registry.settings['facebook.app_id']
    secret_key = request.registry.settings['facebook.secret_key']
    try:
        signed_request = SignedRequest.parse(signed_request,
                                             secret_key)
    except SignedRequestError as exc:
        log.exception('invalid signed request')
        raise HTTPBadRequest(exc)

    payment_id = unicode(signed_request['payment_id'])
    if payment_id != request.matchdict['order_id']:
        log.error(
            'mismatch between order_id=%r in URI and payment_id=%r in'
            ' signed request', request.matchdict['order_id'], payment_id)
        raise HTTPBadRequest('invalid payment ID')

    token = get_application_access_token(app_id, secret_key)
    # TODO use facepy.GraphAPI when the fix for Decimal is released
    # https://github.com/jgorset/facepy/issues/86
    payment = requests.get('https://graph.facebook.com/' + payment_id,
                           params={'access_token': token})
    payment = payment.json(parse_float=Decimal)

    notify_order_received(request, payment)


@subscriber(ChangeNotification, update_object='payments')
def process_payments(event):
    request = event.request
    entries = request.json_body['entry']
    batch = [{'relative_url': entry['id'], 'method': 'GET'}
             for entry in entries]
    log.debug('Posting batch batch=%s', batch)
    try:
        results = request.graph_api.post('/', batch=json.dumps(batch))
    except FacepyError:
        log.exception('Batch request failed. json_body=%s', request.json_body)
        raise HTTPFailedDependency('Facebook batch request failed')

    for result in results:
        try:
            payment = json.loads(result['body'], parse_float=Decimal)
        except (KeyError, ValueError):
            log.exception('Facebook batch response invalid: results=%s',
                          results)
            raise HTTPFailedDependency('Facebook batch response invalid')

        notify_order_received(request, payment)


def notify_order_received(request, payment):
    log.debug('notifying OrderReceived: payment=%s', payment)
    try:
        request.registry.notify(OrderReceived(request, payment))
    except Exception:
        log.exception(
            u'notifying OrderReceived: user_id=%s, order_id=%s, actions=%s',
            payment[u'user'][u'id'],
            payment[u'id'],
            payment[u'actions'],
        )
        try:
            request.registry.notify(OrderProcessingError(request, payment))
        except Exception:
            log.critical(
                u'notifying OrderProcessingError:  user_id=%s, order_id=%s '
                u'actions=%s',
                payment[u'user'][u'id'],
                payment[u'id'],
                payment[u'actions'],
                exc_info=True,
            )
        raise HTTPServerError()



# -*- coding: utf-8 -*-
import unittest

import mock

from facepy import SignedRequest


class TestLib(unittest.TestCase):

    def test_base(self):
        from pyramid_facebook.lib import Base
        req = mock.Mock()
        ctx = mock.Mock()

        b = Base(ctx, req)
        self.assertEqual(ctx, b.context)
        self.assertEqual(req, b.request)

    def test_encrypt_decrypt_request(self):
        from pyramid_facebook.lib import encrypt_signed_request
        data = {
            u'issued_at': 1297110048,
            u'user': {
                u'locale': u'en_US',
                u'country': u'ca'
                },
            u'algorithm': u'HMAC-SHA256'
            }

        req = encrypt_signed_request('secret key', data)

        res = SignedRequest.parse(req, 'secret key')
        self.assertEqual(data.keys(), res.keys())
        self.assertEqual(data, res)



import unittest

import mock


settings = """[server:main]
use = egg:waitress#main
host = 0.0.0.0
port = 5000
[app:main]
use = call:%s:main
facebook.app_id = 123
facebook.secret_key = abcdef1234567890
facebook.namespace = test_canvas
facebook.scope =
facebook.rt-subscriptions =
    user = name, friends
    payments = actions, disputes""" % __name__


def main(global_config, **settings):
    from pyramid.config import Configurator
    config = Configurator(settings=settings)
    config.include('pyramid_facebook')
    return config.make_wsgi_app()


class TestManageSubscriptions(unittest.TestCase):

    def _get_command(self, *params):
        from pyramid_facebook.real_time import ManageSubscriptions
        return ManageSubscriptions(params)

    def _get_subscriptions_list(self):
        return {
            'data': [
                {'active': True,
                 'fields': ['friends', 'name'],
                 'object': 'user',
                 'callback_url': 'http://example.com/test_canvas/real-time',
                 },
                {'active': True,
                 'fields': ['actions', 'disputes'],
                 'object': 'payments',
                 'callback_url': 'http://example.com/test_canvas/real-time'
                 },
            ]}

    def setUp(self):
        self.patch_open = mock.patch('paste.deploy.loadwsgi.open',
                                     mock_open(read_data=settings),
                                     create=True)
        self.patch_open.start()

        self.patch_graph_api = mock.patch('pyramid_facebook.real_time.'
                                          'GraphAPI')

        self.patch_get_app_token = mock.patch('pyramid_facebook.real_time.'
                                              'get_application_access_token')

        self.patch_list = mock.patch('pyramid_facebook.real_time.'
                                     'list_real_time_subscriptions')

        self.m_graph_api = self.patch_graph_api.start()

        self.m_list = self.patch_list.start()

        self.m_list.return_value = self._get_subscriptions_list()

        self.patch_get_app_token.start().return_value = 'app_token'

    def tearDown(self):
        self.patch_open.stop()
        self.patch_graph_api.stop()
        self.patch_list.stop()
        self.patch_get_app_token.stop()

    def test_list(self):
        ms = self._get_command('fake.ini', 'list')
        ms.run()

    def test_update(self):
        ms = self._get_command('fake.ini', 'update', 'user=name,friends')
        ms.run()

    def test_delete(self):
        ms = self._get_command('fake.ini', 'delete')
        ms.run()

    def test_setup(self):
        ms = self._get_command('fake.ini', 'setup')
        ms.run()

    @mock.patch('pyramid_facebook.real_time.sys')
    def test_command_line_script(self, m_sys):
        from pyramid_facebook.real_time import command_line_script
        m_sys.argv = ['pfacebook-real-time', 'fake.ini', 'list']
        command_line_script()


def mock_open(read_data):
    """Mock module does not mock readline and readlines method on file object.
    Unfortunately config parser called when boostraping the app uses readline.
    """
    mock_obj = mock.MagicMock(name='open', spec=open)
    handle = mock.MagicMock(spec=file)
    handle.__enter__.return_value = handle
    handle.read.return_value = read_data

    class ReadLineMock():
        def __init__(self, read_data):
            self.read_data = read_data.split('\n')

        def next(self):
            if self.read_data:
                return self.read_data.pop(0)
            return ''

    handle.readline.side_effect = ReadLineMock(read_data)
    mock_obj.return_value = handle
    return mock_obj



from __future__ import absolute_import

from pyramid_facebook.lib import js_redirect_tpl
from pyramid_facebook.tests.functional import TestView


class TestCanvas(TestView):
    def test_redirect_to_app(self):
        from pyramid_facebook.tests import conf
        namespace = conf["facebook.namespace"]

        res = self.app.get("/%s/" % namespace, status=200)
        self.assertEqual(
            js_redirect_tpl % {"location": "http://apps.facebook.com/%s" % namespace},
            res.body
        )



import hashlib
import hmac
import json
import logging

from facepy import (
    FacepyError,
    get_application_access_token,
    GraphAPI,
    SignedRequest,
    )
from facepy.exceptions import SignedRequestError

from pyramid.decorator import reify
from pyramid.security import Allow

from pyramid_contextauth import (
    authentication_policy,
    get_authentication_policy,
)

from pyramid_facebook.events import UserSignedRequestParsed

log = logging.getLogger(__name__)

RETRY_NB = 3

ViewCanvas = u'view_canvas'
Authenticate = u'authenticate'
UpdateSubscription = u'update_subscription'
NotifyRealTimeChanges = u'notify_real_time_changes'

FacebookUser = u'facebook-user'
RegisteredUser = u'registered-user'
AdminUser = u'admin-user'
# real time updates https://developers.facebook.com/docs/reference/api/realtime
XHubSigned = u'x-hub-signed'


def includeme(config):
    try:
        config.get_authentication_policy()
    except AttributeError:
        config.include('pyramid_contextauth')
        app_id = config.registry.settings['facebook.app_id']
        secret_key = config.registry.settings['facebook.secret_key']

        policy = get_authentication_policy(config)
        # by default, having a user id means that user is registered
        policy.callback = lambda user_id, request: [RegisteredUser]

        admin_policy = AdminAuthenticationPolicy(app_id, secret_key)

        config.register_authentication_policy(admin_policy, AdminContext)
        config.scan(__name__)
        config.commit()


class SignedRequestContext(object):
    """Security context for facebook signed request routes.

    Attributes: *facebook_data* is information decoded from the signed
    request; *custom_data* can be set by authentication policies that
    build on top of pyramid-facebook to use a central database for users,
    for example.

    The properties *user* and *user_country* are shortcuts for fields
    in *facebook_data*
    """

    __acl__ = [
        (Allow, FacebookUser, Authenticate),
        (Allow, RegisteredUser, ViewCanvas),
        ]

    def __init__(self, request):
        self.request = request
        self.facebook_data = None
        self.custom_data = None

    @reify
    def user(self):
        return self.facebook_data['user']

    @reify
    def user_country(self):
        return self.facebook_data['user']['country']

    def __repr__(self):
        return '<%s facebook_data=%r>' % (
            self.__class__.__name__,
            self.facebook_data
            )


class FacebookCreditsContext(SignedRequestContext):
    "Context for facebook credits callback requests."

    @reify
    def order_details(self):
        """Order details received in `facebook credits callback for payment
        status updates <http://developers.facebook.com/docs/credits/callback/
        #payments_status_update>`_."""
        return json.loads(
            self.facebook_data['credits']['order_details']
            )

    @reify
    def order_info(self):
        """Order info being the order information passed when the `FB.ui method
        <http://developers.facebook.com/docs/reference/javascript/FB.ui/>`_
        is invoked."""
        return self.facebook_data["credits"]["order_info"]

    @reify
    def earned_currency_data(self):
        """Modified field received in `facebook credits callback for payment
        status update for earned app currency
        <http://developers.facebook.com/docs/credits/callback/
        #payments_status_update_earn_app_currency>`_."""
        data = self.item['data']
        if data:
            try:
                data = json.loads(data)
                data = data['modified'] if 'modified' in data else None
            except Exception:
                data = None
        return data

    @reify
    def item(self):
        """The item info as passed when `FB.ui method
        <http://developers.facebook.com/docs/reference/javascript/FB.ui/>`_
        is invoked."""
        return self.order_details['items'][0]


class AccessTokenContext(object):

    def __init__(self, request):
        self.request = request
        self._user_dict = {}
        self._user_id = None

    @property
    def user_dict(self):
        return self._user_dict

    @user_dict.setter
    def user_dict(self, value):
        self._user_dict = value

    @property
    def user_id(self):
        return self._user_id

    @user_id.setter
    def user_id(self, value):
        self._user_id = value


class AdminContext(AccessTokenContext):
    """Context which defines principals as administrative roles user has on
    facebook.

    https://developers.facebook.com/docs/reference/api/application/#roles
    """

    def __init__(self, request):
        super(AdminContext, self).__init__(request)
        self.__acl__ = (
            (Allow, u'group:administrators', UpdateSubscription),
            )


class RealTimeNotificationContext(object):
    "Context for real-time changes notification route."

    __acl__ = (
        (Allow, XHubSigned, NotifyRealTimeChanges),
        )

    def __init__(self, request):
        self.request = request


@authentication_policy(contexts=(SignedRequestContext, FacebookCreditsContext))
class SignedRequestAuthenticationPolicy(object):

    def unauthenticated_userid(self, request):
        context = request.context
        if 'signed_request' not in request.params:
            return None
        context.facebook_data = None
        try:
            context.facebook_data = SignedRequest.parse(
                request.params[u'signed_request'],
                request.registry.settings[u'facebook.secret_key'],
                )
        except SignedRequestError:
            log.warn('SignedRequestError with signature: %s',
                     request.params['signed_request'], exc_info=True)
            return None
        try:
            user_id = int(context.facebook_data[u'user_id'])
        except KeyError:
            # user_id not in facebook_data => user has not authorized app
            log.debug('User has not authorized app.')
        except ValueError:
            log.warn('Invalid user id %r', context.facebook_data[u'user_id'])
        else:
            request.registry.notify(
                UserSignedRequestParsed(request.context, request)
                )
            return user_id
        return None

    def effective_principals(self, request):
        try:
            if request.context.facebook_data['user_id']:
                return [FacebookUser]
        except Exception:
            return []


@authentication_policy(contexts=AccessTokenContext)
class AccessTokenAuthenticationPolicy(object):

    def unauthenticated_userid(self, request):
        try:
            access_token = request.params['access_token']
        except KeyError:
            return None
        request.context.access_token = access_token
        api = GraphAPI(access_token)
        try:
            request.context.user_dict = api.get('me', retry=RETRY_NB)
        except FacepyError:
            log.warn('Authentication failed', exc_info=True)
        try:
            request.context.user_id = long(request.context.user_dict['id'])
            return request.context.user_id
        except KeyError:
            return None
        except ValueError:
            log.warn('Malformated Facebook response', exc_info=True)
        return None

    def effective_principals(self, request):
        p = []
        if request.context.user_id:
            p.append(FacebookUser)
        return p


@authentication_policy(contexts=RealTimeNotificationContext)
class RealTimeNotifAuthenticationPolicy(object):

    def effective_principals(self, request):
        # route predicates already check presence of X-Hub-Signature header
        sig = request.headers[u'X-Hub-Signature']
        verif = hmac.new(
            request.registry.settings['facebook.secret_key'],
            request.body,
            hashlib.sha1
            ).hexdigest()
        if sig == ('sha1=%s' % verif):
            return [XHubSigned]
        else:
            log.warn('X-Hub-Signature invalid - expected %s, received %s',
                     verif, sig)
        return []


class AdminAuthenticationPolicy(AccessTokenAuthenticationPolicy):
    """Check from on facebook what role the user has on the application via its
    access_token. Role the user should assume, one of 'group:administrators',
    'group:developers', 'group:testers' or 'group:insights users'

    https://developers.facebook.com/docs/reference/api/application/#roles
    """
    def __init__(self, app_id, secret_key):
        self.app_id = app_id
        self.secret_key = secret_key

    @reify
    def app_token(self):
        return get_application_access_token(self.app_id, self.secret_key)

    def effective_principals(self, request):
        p = super(AdminAuthenticationPolicy,
                  self).effective_principals(request)

        api = GraphAPI(self.app_token)
        try:
            roles = api.get('%s/roles' % self.app_id, retry=RETRY_NB)
        except FacepyError:
            log.warn('Get roles failed', exc_info=True)
            return p
        for role in roles['data']:
            if request.context.user_id == long(role['user']):
                p.append('group:' + role['role'])
        return p



import posixpath

from zope.interface import Interface, Attribute
from pyramid.view import view_config
from pyramid.traversal import resource_path_tuple
from pyramid.httpexceptions import HTTPMethodNotAllowed


def includeme(config):
    config.add_directive('add_opengraph_collection', add_collection)
    config.scan('.opengraph')


class IResource(Interface):
    """Interface used to configure a view for all OpenGraph objects."""

    prefixes = Attribute('Extra OpenGraph prefixes')
    properties = Attribute('Dictionary of properties-values')


def add_collection(config, cls, pattern):
    """Add route and resource factory for a collection of objects.

    Interface for the collection:

        __name__ and __parent__ attributes set to None

        __init__(self, request)

        __getitem__(self, key)
            Returns an object implementing IResource.
    """
    # note that cls.__name__ is the Python class attribute, e.g.
    # 'VirtualGoodCollection', not the Pyramid instance attribute
    # that is used for URI generation, e.g. 'packs'
    namespace = config.registry.settings['facebook.namespace']
    pattern = posixpath.join(namespace, pattern, '{object_id}')

    config.add_route(name='opengraph_%s' % cls.__name__.lower(),
                     pattern=pattern,
                     traverse='/{object_id}',
                     factory=cls,
                     use_global_views=True)


@view_config(context=IResource,
             request_method='GET',
             renderer='pyramid_facebook:templates/opengraph_object.mako')
def show_opengraph_object(context, request):
    # TODO support redirect_url
    resource_path = resource_path_tuple(context)[1:]
    og_url = request.current_route_url(traverse=resource_path,
                                       _query=request.GET)
    # TODO make the update conditional
    context.properties.update({
        'fb:app_id': request.registry.settings['facebook.app_id'],
        'og:url': og_url,
    })
    return {'prefixes': context.prefixes,
            'properties': context.properties}


@view_config(context=IResource)
def default_opengraph_object(context, request):
    raise HTTPMethodNotAllowed()



from decimal import Decimal
import json


class _number_str(float):
    # kludge to have decimals correctly output as JSON numbers;
    # converting them to strings would cause extra quotes

    def __init__(self, o):
        self.o = o

    def __repr__(self):
        return str(self.o)


class JSONEncoder(json.JSONEncoder):
    """Custom encoder that supports Decimal.
    """
    def default(self, o):

        if isinstance(o, Decimal):
            return _number_str(o)

        return super(JSONEncoder, self).default(o)  # pragma: no cover



# -*- coding: utf-8 -*-
import setuptools

# this mixes setup deps with test deps, but our internal test tool
# (based on nosetests) was not working properly with test deps
setup_requires = [
    'd2to1',
]

setuptools.setup(
    setup_requires=setup_requires,
    d2to1=True,
    test_suite='nose.collector',
    entry_points="""
        [console_scripts]

        pfacebook-real-time=pyramid_facebook.real_time:command_line_script
    """
)



import json
import hashlib
import hmac
from decimal import Decimal

import mock
from facepy import FacepyError
from pyramid.config import Configurator
from pyramid.decorator import reify
from webtest import TestApp

from . import JSONEncoder


class Mixin(object):
    """A mixin integratiom test for real-time update on payments.
    """
    user_id = 700
    application_id = 2222
    virtual_good_id = 1

    @reify
    def config(self):
        self.addCleanup(delattr, self, 'config')
        return Configurator(settings={})

    @reify
    def app(self):
        self.addCleanup(delattr, self, 'app')
        return TestApp(self.config.make_wsgi_app())

    @property
    def namespace(self):
        return self.config.registry.settings['facebook.namespace']

    @property
    def real_time_path(self):
        return ('/%s/real-time' % self.namespace)

    def sign(self, body):
        secret_key = self.config.registry.settings['facebook.secret_key']
        return "sha1=%s" % hmac.new(secret_key, body, hashlib.sha1).hexdigest()

    def get_payment_dict(self):
        return {
            u'actions': [{
                u'amount': u'39.99',
                u'currency': u'EUR',
                u'status': u'completed',
                u'time_created': u'2013-08-08T21:19:59+0000',
                u'time_updated': u'2013-08-08T21:19:59+0000',
                u'type': u'charge',
                }],
            u'application': {
                u'id': self.application_id,
                u'name': u'TEST',
                u'namespace': u'%s' % self.namespace
                },
            u'country': u'CA',
            u'created_time': u'2013-08-08T21:19:59+0000',
            u'id': u'111',
            u'items': [{
                u'product': u'http://example.com/%s/item/%s' % (
                    self.namespace,
                    self.virtual_good_id,
                    ),
                u'quantity': 1,
                u'type': u'IN_APP_PURCHASE',
                }],
            u'payout_foreign_exchange_rate': Decimal('1.3131459'),
            u'refundable_amount': {u'amount': u'39.99', u'currency': u'EUR'},
            u'test': 1,
            u'user': {
                u'id': self.user_id,
                u'name': u'Joe Manix'
                }
            }

    def get_notification_dict(self):
        return {
            u'object': u'payments',
            u'entry': [{
                u'id': u'111',
                u'time': 1347996346,
                u'changed_fields': ['actions']
                }]
            }

    @mock.patch('pyramid_facebook.utility.get_application_graph_api')
    def test_real_time_payment_update_completed(self, m_get_graph_api):
        from pyramid_facebook.events import OrderReceived
        json_body = json.dumps(self.get_notification_dict())

        or_subscriber = mock.Mock()
        self.config.add_subscriber(or_subscriber, OrderReceived)

        payment = self.get_payment_dict()

        batch_results = [{u'body': json.dumps(payment, cls=JSONEncoder)}]

        m_graph_api = m_get_graph_api.return_value
        m_graph_api.post.return_value = batch_results

        self.app.post(
            self.real_time_path,
            json_body,
            headers={
                'X-Hub-Signature': self.sign(json_body),
                'Content-Type': 'application/json',
                }
            )

        self.assertEqual(1, len(or_subscriber.call_args_list))
        event = or_subscriber.call_args_list[0][0][0]

        self.assertEqual(payment, event.payment)

    @mock.patch('pyramid_facebook.utility.get_application_graph_api')
    def test_real_time_payment_update_failed(self, m_get_graph_api):
        from pyramid_facebook.events import OrderReceived
        json_body = json.dumps(self.get_notification_dict())

        or_subscriber = mock.Mock()
        self.config.add_subscriber(or_subscriber, OrderReceived)

        payment = self.get_payment_dict()
        payment['actions'][0]['status'] = 'failed'

        batch_results = [{u'body': json.dumps(payment, cls=JSONEncoder)}]

        m_graph_api = m_get_graph_api.return_value
        m_graph_api.post.return_value = batch_results

        self.app.post(
            self.real_time_path,
            json_body,
            headers={
                'X-Hub-Signature': self.sign(json_body),
                'Content-Type': 'application/json',
                }
            )

        self.assertEqual(1, len(or_subscriber.call_args_list))
        event = or_subscriber.call_args_list[0][0][0]

        self.assertEqual(payment, event.payment)

    @mock.patch('pyramid_facebook.utility.get_application_graph_api')
    def test_real_time_payment_with_facebook_fail(self, m_get_graph_api):
        from pyramid_facebook.events import OrderReceived
        json_body = json.dumps(self.get_notification_dict())

        or_subscriber = mock.Mock()
        self.config.add_subscriber(or_subscriber, OrderReceived)

        m_graph_api = m_get_graph_api.return_value
        m_graph_api.post.side_effect = FacepyError('shit hits the fan')

        self.app.post(
            self.real_time_path,
            json_body,
            headers={
                'X-Hub-Signature': self.sign(json_body),
                'Content-Type': 'application/json',
                },
            status=500,
            )

        self.assertEqual(0, or_subscriber.call_count)

    @mock.patch('pyramid_facebook.utility.get_application_graph_api')
    def test_real_time_payment_fb_batch_response_invalid(self,
                                                         m_get_graph_api):
        from pyramid_facebook.events import OrderReceived
        json_body = json.dumps(self.get_notification_dict())

        or_subscriber = mock.Mock()
        self.config.add_subscriber(or_subscriber, OrderReceived)

        batch_results = [{u'body': 'not a valid json body'}]

        m_graph_api = m_get_graph_api.return_value
        m_graph_api.post.return_value = batch_results

        self.app.post(
            self.real_time_path,
            json_body,
            headers={
                'X-Hub-Signature': self.sign(json_body),
                'Content-Type': 'application/json',
                },
            status=500
            )

        self.assertEqual(0, len(or_subscriber.call_args_list))



# -*- coding: utf-8 -*-
import base64
import hashlib
import hmac
import json


js_redirect_tpl = """<html>
  <body>
    <script>
      window.top.location = "%(location)s";
    </script>
  </body>
</html>"""


# TODO mostly useless, delete

class Base(object):
    "Base class for views and events"
    def __init__(self, context, request):
        self.context = context
        self.request = request


def _base64_url_encode(inp):
    """ Facebook base64 decoder based on `Sunil Arora's blog post
    <http://sunilarora.org/parsing-signedrequest-parameter-in-python-bas>`_.

    :param inp: input `str` to be encoded
    :return: `base64` encoded utf8 unicode data
    """
    return unicode(base64.b64encode(inp, '-_').strip('=').encode('utf8'))


# TODO deprecate in favor of facepy

def encrypt_signed_request(secret_key, data):
    """Encrypts data the way facebook does for permit testing. Adds algorithm
    key to dict.

    :param secret_key: Facebook application' secret key.
    :param data: a dictionary of data to sign.
    :return: Signed request as defined by `Facebook documentation
             <http://developers.facebook.com/docs/authentication/
             signed_request/>`_

    """
    data = data.copy()
    data.update(algorithm='HMAC-SHA256')

    payload = _base64_url_encode(json.dumps(data)).encode('utf8')
    signature = _base64_url_encode(
        hmac.new(
            secret_key.encode('utf8'),
            msg=payload.encode('utf8'),
            digestmod=hashlib.sha256
            ).digest()
        )
    return '%s.%s' % (signature, payload)



import unittest

import mock


class TestPredicate(unittest.TestCase):
    def test_request_params_predicate(self):
        from pyramid_facebook.predicates import request_params_predicate

        predicate = request_params_predicate('a', 'b', 'c', d=123, e='yo')

        request = mock.Mock()
        request.params = dict(a=1, b=2, c=3, d=123, e='yo')
        self.assertTrue(predicate(None, request))

        request.params = dict(a=1, b=2)
        self.assertFalse(predicate(None, request))

        request.params = dict(a=1, b=2, c=1, d=321, e='oy')
        self.assertFalse(predicate(None, request))

    def test_nor_predicate(self):
        from pyramid_facebook.predicates import nor_predicate
        predicate = nor_predicate(test=('a', 'b', 'c'))

        request = mock.Mock()
        request.params = dict(test='a')

        self.assertTrue(predicate(None, request))

        request.params = dict(test='c')
        self.assertTrue(predicate(None, request))

        request.params = dict(not_test='c')
        self.assertFalse(predicate(None, request))

        request.params = dict(test='d')
        self.assertFalse(predicate(None, request))

    def test_headers_predicate(self):
        from pyramid_facebook.predicates import headers_predicate

        predicate = headers_predicate('X-header-1', **{'X-header-2': 'value'})

        request = mock.Mock()
        request.headers = {}

        self.assertFalse(predicate(None, request))

        request.headers = {'X-header-1': 'yo'}

        self.assertFalse(predicate(None, request))

        request.headers = {'X-header-1': 'yo', 'X-header-2': 'not value'}

        self.assertFalse(predicate(None, request))

        request.headers = {'X-header-1': 'yo', 'X-header-2': 'value'}

        self.assertTrue(predicate(None, request))


class TestMatchedRoutePredicate(unittest.TestCase):

    def test_predicate_matched(self):
        from pyramid_facebook.predicates import MatchedRouteEventPredicate

        pred = MatchedRouteEventPredicate(u'route_name', None)

        event = mock.Mock()
        event.request.matched_route.name = u'route_name'

        self.assertTrue(pred(event))

    def test_predicate_unmatched(self):
        from pyramid_facebook.predicates import MatchedRouteEventPredicate

        pred = MatchedRouteEventPredicate(u'route_name', None)

        event = mock.Mock()
        event.request.matched_route.name = u'not_same_route_name'

        self.assertFalse(pred(event))

    def test_text(self):
        from pyramid_facebook.predicates import MatchedRouteEventPredicate

        pred = MatchedRouteEventPredicate(u'route_name', None)

        self.assertEqual(u'route_to_match = route_name', pred.text())


class TestContextEcentPredicate(unittest.TestCase):

    def test_predicate_matched(self):
        from pyramid_facebook.predicates import ContextEventPredicate
        from pyramid_facebook.security import SignedRequestContext

        event = mock.Mock()
        event.request.context = SignedRequestContext(None)

        pred = ContextEventPredicate(SignedRequestContext, None)

        self.assertTrue(pred(event))

    def test_predicate_unmatched(self):
        from pyramid_facebook.predicates import ContextEventPredicate
        from pyramid_facebook.security import (
            SignedRequestContext,
            RealTimeNotificationContext,
            )

        event = mock.Mock()
        event.request.context = RealTimeNotificationContext(None)

        pred = ContextEventPredicate(SignedRequestContext, None)

        self.assertFalse(pred(event))

    def test_text(self):
        from pyramid_facebook.predicates import ContextEventPredicate
        from pyramid_facebook.security import SignedRequestContext

        pred = ContextEventPredicate(SignedRequestContext, None)

        self.assertEqual(
            u"expected_context_cls = <class "
            "'pyramid_facebook.security.SignedRequestContext'>",
            pred.text()
            )


class TestPermissionEventPredicate(unittest.TestCase):

    def test_call(self):
        from pyramid_facebook.predicates import PermissionEventPredicate

        pred = PermissionEventPredicate('view-canvas', None)

        pred(mock.Mock())

    def test_text(self):
        from pyramid_facebook.predicates import PermissionEventPredicate
        pred = PermissionEventPredicate('view-canvas', None)

        self.assertEqual(u"permission = 'view-canvas'", pred.text())



import json
import hashlib
import hmac

import mock
from pyramid.config import Configurator
from pyramid.decorator import reify
from pyramid.httpexceptions import HTTPForbidden
from webtest import TestApp


class Mixin(object):

    @reify
    def config(self):
        self.addCleanup(delattr, self, 'config')
        return Configurator(settings={})

    @reify
    def app(self):
        self.addCleanup(delattr, self, 'app')
        return TestApp(self.config.make_wsgi_app())

    @property
    def real_time_path(self):
        return ('/%s/real-time' %
                self.config.registry.settings['facebook.namespace'])

    def sign(self, body):
        secret_key = self.config.registry.settings['facebook.secret_key']
        return "sha1=%s" % hmac.new(secret_key, body, hashlib.sha1).hexdigest()

    def get_update_dict(self):
        return {
            "object": "user",
            "entry": [
                {"uid": 111,
                 "changed_fields": ["name", "picture"],
                 "time": 123456,
                 },
                {"uid": 222,
                 "changed_fields": ["friends"],
                 "time": 123456,
                 },
                ]
            }

    def test_real_time_update_invalid_signature(self):
        with self.assertRaises(HTTPForbidden):
            self.app.post(
                self.real_time_path,
                headers={
                    'X-Hub-Signature': 'invalid signature',
                    'Content-Type': 'application/json',
                    }
                )

    def test_real_time_update(self):
        from pyramid_facebook.events import ChangeNotification
        json_body = json.dumps(self.get_update_dict())

        rt_subscriber = mock.Mock()
        self.config.add_subscriber(rt_subscriber, ChangeNotification)

        self.app.post(
            self.real_time_path,
            json_body,
            headers={
                'X-Hub-Signature': self.sign(json_body),
                'Content-Type': 'application/json',
                }
            )

        self.assertEqual(1, len(rt_subscriber.call_args_list))
        event = rt_subscriber.call_args_list[0][0][0]

        self.assertEqual(self.get_update_dict(), event.json_body)
